package com.taobao.movie.android.app.oscar.ui.homepage.item

import android.animation.Animator
import android.animation.AnimatorListenerAdapter
import android.graphics.Bitmap
import android.support.v7.widget.RecyclerView
import android.text.TextUtils
import android.view.View
import android.view.ViewGroup
import android.widget.ImageView
import com.airbnb.lottie.LottieAnimationView
import com.airbnb.lottie.LottieComposition
import com.alibaba.fastjson.JSON
import com.taobao.movie.android.app.oscar.ui.homepage.util.HomeLottieSwitchHelper
import com.taobao.movie.android.app.oscar.ui.widget.SafeLottieAnimationView
import com.taobao.movie.android.common.scheme.MovieNavigator
import com.taobao.movie.android.commonutil.kotlin.print
import com.taobao.movie.android.home.R
import com.taobao.movie.android.integration.oscar.model.BannerMo
import com.taobao.movie.android.integration.oscar.model.HomeBreadLottie
import com.taobao.movie.android.sdk.infrastructure.monitor.business.LottieMonitorPoint
import com.taobao.movie.android.sdk.infrastructure.monitor.business.MonitorPointConstant
import com.taobao.movie.android.sdk.infrastructure.usertrack.UTFacade
import com.taobao.movie.appinfo.MovieAppInfo
import com.taobao.movie.appinfo.util.DisplayUtil

/**
 * fake重试最大次数
 */
internal const val MAX_FAKE_TIMES = 30

/**
 * 氛围 banner 帮助类
 *
 * @author yidong.yyd
 * @since 2019-12-18
 */
class HomeEnvironmentBannerHelper(private val bannerMo: BannerMo,
                                  private val itemView: View,
                                  private val cityCode: String,
                                  private val lottieDownloadCallback: ((url: String) -> Unit))
    : View.OnClickListener, RecyclerView.OnChildAttachStateChangeListener {

    /* 氛围banner的图片展示 */
    private val environmentImageView: ImageView = itemView.findViewById(R.id.iv_environment_banner)
    /* 氛围banner的点击holder */
    private val environmentClickHolderView: View = itemView.findViewById(R.id.view_environment_banner_click_holder)
    /* 氛围banner的填充颜色 */
    private val environmentFillHolderView: View = itemView.findViewById(R.id.view_environment_banner_holder)
    /* 氛围banner的lottie */
    private val environmentLottieView: LottieAnimationView = itemView.findViewById(R.id.lottie_environment_banner)
    /* 伪装 */
    private val environmentBannerFakerView: ImageView = itemView.findViewById(R.id.iv_environment_banner_faker)

    /* lottie的网络地址 */
    var lottieUrl: String? = null
    /* lottie下载失败的情况下走图片兜底 */
    private var useLottie = true
    /* 在 RecyclerView 中使用 Lottie */
    private val lottieResourceRecycler: LottieResourceRecycler
    /* 是否塞上了 bitmap */
    private var isFakerOwnBitmap = false
    /* 滑动状态 */
    private var scrollState: Int = RecyclerView.SCROLL_STATE_IDLE
    /* 是否在 Window 上了 */
    private var isAttachToWindow = true
    /* 首页宣发视频是否在播放 */
    var isHomeBannerVideoPlay = false
    /* fake 重试次数 */
    private var cacheFakeBitmapTimes = 0

    init {
        // 默认开启
        useLottie = HomeLottieSwitchHelper.instance.isHomeEnvBannerLottie()
        lottieResourceRecycler = LottieResourceRecycler(environmentLottieView) { lottieFail(it) }
        environmentClickHolderView.setOnClickListener(this)
        download()
        UTFacade.setSpmTag(environmentLottieView, "HomeSpringBannerExpose.1")
        UTFacade.setCustomExposureTag(environmentLottieView, "HomeSpringBannerExpose", null,
                "bannerId", bannerMo.id,
                "city", cityCode,
                "adv_code", bannerMo.advertiseContainer,
                "adv_type", "${bannerMo.advertiseType}",
                "index", "1")
        SafeLottieAnimationView.replaceFailureListener(environmentLottieView) {
            lottieFail(MonitorPointConstant.LOTTIE_UNKNOWN)
        }
    }

    /**
     * 氛围 banner 是否正在展示
     *
     * @return `true`: 展示中
     */
    private fun isEnvironmentBannerShowing(): Boolean = environmentClickHolderView.visibility === View.VISIBLE

    /**
     * 滑动状态改变
     */
    fun scrollChanged(newState: Int, firstVisiblePosition: Int, lastVisiblePosition: Int, position: Int) {
        scrollState = newState
        if (!useLottie) return
        if (newState != RecyclerView.SCROLL_STATE_IDLE && isEnvironmentBannerShowing() && isEnvironmentBannerLottieAnimating()) {
            pauseEnvironmentBannerLottie()
        }
        if (newState == RecyclerView.SCROLL_STATE_IDLE && isAttachToWindow && isFakerOwnBitmap) {
            environmentLottieView.visibility = View.VISIBLE
            environmentBannerFakerView.visibility = View.INVISIBLE
        }
        if (newState == RecyclerView.SCROLL_STATE_IDLE && isEnvironmentBannerShowing() && !isEnvironmentBannerLottieAnimating()
                && (firstVisiblePosition <= position || lastVisiblePosition >= position)) {
            continueEnvironmentBannerLottie()
        }
    }

    /**
     * lottie 下载成功
     */
    fun lottieDownloadSuccess(lottieComposition: LottieComposition) {
        lottieUrl?.let { lottieResourceRecycler.unzipResource(it) { unzipSuccess(lottieComposition) } }
    }

    private fun unzipSuccess(lottieComposition: LottieComposition) {
        statistic(MonitorPointConstant.LOTTIE_RESOURCE_DOWNLOAD_SUCCESS)
        showEnvironmentBanner(true)
        //开启硬件加速
        environmentLottieView.useHardwareAcceleration()
        environmentLottieView.setComposition(lottieComposition)
        continueEnvironmentBannerLottie()
        if (fitLottieHeight(false)) {
            statistic(MonitorPointConstant.LOTTIE_RESOURCE_VALID)
        }
    }

    /**
     * lottie 下载失败
     */
    fun lottieDownloadFail() = lottieFail(MonitorPointConstant.LOTTIE_RESOURCE_DOWNLOAD_FAIL)

    /**
     * 页面还原
     */
    fun resume() {
        if (isAttachToWindow && scrollState == RecyclerView.SCROLL_STATE_IDLE) {
            continueEnvironmentBannerLottie()
        }
    }

    /**
     * 页面 pause 了暂停动画
     */
    fun pause() {
        pauseEnvironmentBannerLottie()
    }

    /**
     * 隐藏lottie，然后显示出 faker view
     */
    fun hideLottie() {
        if (isFakerOwnBitmap && useLottie) {
            environmentLottieView.visibility = View.GONE
            environmentBannerFakerView.visibility = View.VISIBLE
        }
    }

    /**
     * 隐藏所有 View
     */
    fun hideAll() {
        if (environmentLottieView.isAnimating) {
            environmentLottieView.cancelAnimation()
        }
        environmentImageView.visibility = View.GONE
        environmentClickHolderView.visibility = View.GONE
        environmentFillHolderView.visibility = View.GONE
        environmentLottieView.visibility = View.GONE
        environmentBannerFakerView.visibility = View.GONE
    }

    override fun onClick(v: View?) {
        UTFacade.buttonClickWithoutPage("HomeSpringBannerClick",
                "bannerId", bannerMo.id,
                "city", cityCode,
                "adv_code", bannerMo.advertiseContainer,
                "adv_type", "${bannerMo.advertiseType}",
                "index", "1")
        MovieNavigator.handleBannerUrl(itemView.context, bannerMo.actionUrl, bannerMo.deeplinkUrl)
    }

    override fun onChildViewDetachedFromWindow(p0: View) {
        isAttachToWindow = false
        if (scrollState != RecyclerView.SCROLL_STATE_IDLE) {
            hideLottie()
        }
    }

    override fun onChildViewAttachedToWindow(p0: View) {
        isAttachToWindow = true
        if (itemView === p0) pauseEnvironmentBannerLottie()
        if (scrollState != RecyclerView.SCROLL_STATE_IDLE) {
            hideLottie()
        }
    }

    /**
     * 下载数据
     */
    private fun download() {
        if (useLottie) downloadLottie() else downloadImage()
    }

    /**
     * 下载lottie
     */
    private fun downloadLottie() {
        try {
            val homeBreadLottie = JSON.parseObject(bannerMo.extensions, HomeBreadLottie::class.java)
            homeBreadLottie?.lottie?.let {
                lottieUrl = it
                statistic(MonitorPointConstant.LOTTIE_RESOURCE_URL)
                val success = HomeEnvironmentLoader.getInstance().setConsumer(it) { lottieComposition, unzippedUrl, isFail ->
                    if (!isFail && !TextUtils.isEmpty(unzippedUrl) && lottieComposition != null) {
                        lottieResourceRecycler.setImageAsset(unzippedUrl)
                        unzipSuccess(lottieComposition)
                    } else {
                        lottieDownloadCallback.invoke(it)
                    }
                }
                if (!success) {
                    lottieDownloadCallback.invoke(it)
                }
            } ?: run { lottieFail(MonitorPointConstant.LOTTIE_RESOURCE_NOT_FOUND) }
        } catch (e: Exception) {
            e.print()
            lottieFail(MonitorPointConstant.LOTTIE_DOWNLOAD_EXCEPTION)
        }
    }

    /**
     * 下载氛围 banner
     */
    private fun downloadImage() {
        MovieAppInfo.getInstance().urlImageViewDownloader.download(itemView.context, bannerMo.bigPicUrl, this::handleImage)
    }

    /**
     * 处理图片
     */
    private fun handleImage(bitmap: Bitmap) {
        environmentImageView.setImageBitmap(bitmap)
        fillColor(bitmap)
        showEnvironmentBanner(false)
        environmentImageView.post { fitSize(bitmap) }
    }

    /**
     * 展示氛围 banner
     */
    private fun showEnvironmentBanner(showLottie: Boolean) {
        environmentClickHolderView.visibility = View.VISIBLE
        environmentFillHolderView.visibility = View.VISIBLE
        environmentClickHolderView.layoutParams?.let {
            val displayWidth = DisplayUtil.getDisplayWidthInPx()
            if (displayWidth == 0) { // 这个可能返回 0
                return
            }
            it.height = displayWidth * 145 / 360
        }
        if (showLottie) {
            environmentBannerFakerView.visibility = View.INVISIBLE
            environmentLottieView.visibility = View.VISIBLE
            environmentLottieView.addAnimatorListener(object : AnimatorListenerAdapter() {
                override fun onAnimationStart(animation: Animator) {
                    statistic(MonitorPointConstant.LOTTIE_RESOURCE_PLAY)
                    environmentLottieView.post { fitLottieHeight() }
                    environmentLottieView.removeAnimatorListener(this)
                }
            })
            environmentImageView.visibility = View.GONE
        } else {
            environmentLottieView.visibility = View.GONE
            environmentBannerFakerView.visibility = View.GONE
            environmentImageView.visibility = View.VISIBLE
        }
    }

    /**
     * 处理Lottie的高度
     */
    private fun fitLottieHeight(failRetry: Boolean = true): Boolean {
        if (environmentLottieView.drawable?.intrinsicHeight ?: 0 == 0 || environmentLottieView.drawable?.intrinsicWidth ?: 0 == 0) {
            if (failRetry) {
                lottieFail(MonitorPointConstant.LOTTIE_RESOURCE_INVALID)
            }
            return false
        }
        val dp277 = DisplayUtil.dp2pxInt(277f)
        // 应该的长度
        val shouldHeight = if (dp277 > itemView.height) dp277 else itemView.height
        // 但是展示出来的长度
        val lottieShouldHeight = (environmentLottieView.drawable?.intrinsicHeight ?: 0) *
                itemView.width / (environmentLottieView.drawable?.intrinsicWidth ?: 1)
        if (lottieShouldHeight > shouldHeight) {
            val layoutParams = environmentLottieView.layoutParams as ViewGroup.MarginLayoutParams
            // 如果已经设置的值跟算出来的值一样的话不重新设置
            if (layoutParams.height != lottieShouldHeight || layoutParams.topMargin != (shouldHeight - lottieShouldHeight)) {
                layoutParams.height = lottieShouldHeight
                layoutParams.topMargin = shouldHeight - lottieShouldHeight
                environmentLottieView.layoutParams = layoutParams

                val holderLayoutParams = environmentBannerFakerView.layoutParams as ViewGroup.MarginLayoutParams
                holderLayoutParams.height = lottieShouldHeight
                holderLayoutParams.topMargin = shouldHeight - lottieShouldHeight
                environmentBannerFakerView.layoutParams = holderLayoutParams
            }
        } else {
            // 可能会有空缺，拿 environmentFillHolderView 填充
            val bitmap = cacheBitmap(environmentLottieView)
            fillColor(bitmap)
            fitSize(bitmap)
        }
        fake()
        return true
    }

    /**
     * post 一下是因为上一步操作可能是在 requestLayout， 等测量好了再拿数据
     */
    private fun fake() {
        environmentLottieView.post {
            cacheBitmap(environmentLottieView)?.let {
                if (it.width != environmentLottieView.width || it.height != environmentLottieView.height) {
                    cacheFakeBitmapTimes++
                    if (cacheFakeBitmapTimes > MAX_FAKE_TIMES) {
                        lottieFail(MonitorPointConstant.LOTTIE_FAKE_TIMEOUT)
                    } else {
                        fake()
                    }
                } else {
                    isFakerOwnBitmap = true
                    environmentBannerFakerView.setImageBitmap(Bitmap.createBitmap(it))
                }
            } ?: kotlin.run { lottieFail(MonitorPointConstant.LOTTIE_FAKE_FAIL) }
        }
    }

    /**
     * 将 lottie view 转成 bitmap
     *
     * @param view 要转的 View
     * @return
     */
    private fun cacheBitmap(view: View?) = try {
        view?.buildDrawingCache()
        view?.drawingCache ?: null
    } catch (e: Throwable) {
        e.print()
        null
    }

    /**
     * 适配大小，防止过度绘制
     *
     * @param b 将展示的bitmap
     */
    private fun fitSize(b: Bitmap?) {
        environmentImageView.scaleType = ImageView.ScaleType.FIT_XY
//        if (!validBitmap(b)) return
//        val bitmap = b ?: return
//        val width = if (useLottie) environmentLottieView.width else environmentImageView.width
//        val pictureHeight = bitmap.height * width / bitmap.width
//        val delta = itemView.height - pictureHeight
//        if (delta < 0) {
//            environmentImageView.scaleType = ImageView.ScaleType.FIT_XY
//            environmentFillHolderView.visibility = View.GONE
//        } else {
//            val layoutParams = environmentFillHolderView.layoutParams
//            layoutParams.height = delta
//            environmentFillHolderView.layoutParams = layoutParams
//        }
    }

    /**
     * 将其余地方扩充颜色
     *
     * @param b 将展示的bitmap
     */
    private fun fillColor(b: Bitmap?) {
        if (!validBitmap(b)) return
        val bitmap = b ?: return
        try {
            environmentFillHolderView.setBackgroundColor(bitmap.getPixel(1, 1))
        } catch (t: Throwable) {
            t.print()
            lottieFail(MonitorPointConstant.LOTTIE_FILL_COLOR_ERROR)
        }
    }

    /**
     * bitmap 是否可用
     */
    private fun validBitmap(bitmap: Bitmap?): Boolean = bitmap != null && !bitmap.isRecycled && bitmap.width > 0 && bitmap.height > 0

    /**
     * lottie 处理失败
     */
    private fun lottieFail(status: Int) {
        statistic(status)
        useLottie = false
        isFakerOwnBitmap = false
        download()
    }

    /**
     * 氛围 banner lottie 是否正在展示
     *
     * @return `true`: 展示中
     */
    private fun isEnvironmentBannerLottieShowing(): Boolean = environmentLottieView.visibility == View.VISIBLE

    /**
     * lottie 是否在动画
     *
     * @return `true`:正在动画
     */
    private fun isEnvironmentBannerLottieAnimating(): Boolean = environmentLottieView.isAnimating

    /**
     * 暂停动画
     */
    private fun pauseEnvironmentBannerLottie() {
        if (isEnvironmentBannerLottieShowing() && isEnvironmentBannerLottieAnimating()) {
            environmentLottieView.cancelAnimation()
        }
    }

    /**
     * 继续动画
     */
    private fun continueEnvironmentBannerLottie() {
        if (!isEnvironmentBannerLottieShowing() && useLottie && isFakerOwnBitmap) {
            environmentLottieView.visibility = View.VISIBLE
            environmentBannerFakerView.visibility = View.INVISIBLE
        }
        if (isEnvironmentBannerLottieShowing() && !isEnvironmentBannerLottieAnimating() && !isHomeBannerVideoPlay) {
            environmentLottieView.playAnimation()
        }
    }

    /**
     * 埋点
     */
    private fun statistic(status: Int) {
        LottieMonitorPoint()
                .setSceneName(MonitorPointConstant.SCENE_ENVIRONMENT_BANNER)
                .setStatus(status)
                .release()
    }
}