package com.taobao.movie.android.app.oscar.ui.homepage.item

import android.content.Context
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import com.airbnb.lottie.LottieAnimationView
import com.airbnb.lottie.LottieImageAsset
import com.airbnb.lottie.network.FileExtension
import com.taobao.movie.android.common.theme.SkinFileUtils
import com.taobao.movie.android.commonutil.kotlin.print
import com.taobao.movie.android.sdk.infrastructure.monitor.business.MonitorPointConstant
import io.reactivex.Observable
import io.reactivex.ObservableEmitter
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.schedulers.Schedulers
import java.io.File
import java.util.concurrent.ConcurrentHashMap

/**
 * 在 RecyclerView 中使用 Lottie
 *
 * @author yidong.yyd
 * @since 2019-12-20
 */
class LottieResourceRecycler(private val lottieAnimationView: LottieAnimationView, private val failListener: (reason: Int) -> Unit) {

    companion object {
        /**
         * 修复lottie无缓存的问题
         * {@see com.airbnb.lottie.network.NetworkCache#filenameForUrl(String, FileExtension, boolean)}
         *
         * @param url
         * @param extension
         * @param isTemp
         * @return
         */
        private fun filenameForUrl(url: String, extension: FileExtension, isTemp: Boolean): String =
                "lottie_cache_" + url.replace("\\W+".toRegex(), "") + if (isTemp) extension.extension else ".temp" + extension.extension

        /**
         * 更改名字
         *
         * @param context
         * @param url
         * @param extension
         */
        fun rename(context: Context, url: String, extension: FileExtension): File {
            val cacheFile = File(context.cacheDir, filenameForUrl(url, extension, false))
            val file = File(context.cacheDir, filenameForUrl(url, extension, true))
            if (cacheFile != null && !cacheFile.exists() && file != null && file.exists()) {
                file.renameTo(cacheFile)
            }
            return cacheFile
        }

        /**
         * 前缀
         */
        private fun filePrefixNameForUrl(url: String): String = "lottie_cache_" + url.replace("\\W+".toRegex(), "")

        /**
         * 处理数据
         */
        fun emit(emitter: ObservableEmitter<File>, url: String, context: Context) {
            var cacheZipFile: File? = null
            if (url.toLowerCase().endsWith(".zip")) cacheZipFile = rename(context, url, FileExtension.Zip)
            else if (url.toLowerCase().endsWith(".json")) cacheZipFile = rename(context, url, FileExtension.Json)
            if (cacheZipFile?.exists() == true) {
                val rootUnzipDir = context.externalCacheDir.toString() + File.separator + "lottie" + File.separator + filePrefixNameForUrl(url) + File.separator
                val unzipDir = SkinFileUtils.unzipWithReWriteSuffixIgnoreRepetition(cacheZipFile, rootUnzipDir, "", "")
                if ((unzipDir?.exists() == true) && !emitter.isDisposed) emitter.onNext(unzipDir)
            }
            if (!emitter.isDisposed) {
                emitter.onComplete()
            }
        }
    }

    /* 缓存 bitmap 数据 */
    private val map: ConcurrentHashMap<String, Bitmap> = ConcurrentHashMap()

    /* 是否解析失败过 */
    private var isDecodeError = false

    /**
     * 解压 lottie 的 zip
     */
    fun unzipResource(lottieUrl: String, assetSuccessListener: () -> Unit) {
        Observable.create<File> { emit(it, lottieUrl, lottieAnimationView.context) }
                .subscribeOn(Schedulers.io())
                .observeOn(AndroidSchedulers.mainThread())
                .subscribe({ dir ->
                    lottieAnimationView.setImageAssetDelegate { decodeSafely(dir, it) }
                    assetSuccessListener.invoke()
                }, { failListener.invoke(MonitorPointConstant.LOTTIE_UNZIP_ERROR) })
    }

    /**
     * 设置lottie中图片的路径
     */
    fun setImageAsset(dirPath: String) {
        val dir = File(dirPath)
        lottieAnimationView.setImageAssetDelegate { decodeSafely(dir, it) }
    }

    /**
     * 解码异常就换图片
     */
    private fun decodeSafely(dir: File, it: LottieImageAsset) = if (isDecodeError) {
        createDefaultBitmap() // 出过错就一致给 1*1
    } else {
        val bitmap = decodeBitmap(dir, it) // 资源被清除掉会默认生成 1*1 的图片，如果是这样的会降级
        if (bitmap.width == 1 && bitmap.height == 1 && !isDecodeError) {
            failListener.invoke(MonitorPointConstant.LOTTIE_DECODE_ERROR)
            isDecodeError = true
        }
        bitmap
    }

    /**
     * 处理解码
     *
     * @param unzipDir 解压后的文件夹
     * @param asset [LottieImageAsset]
     * @return 生成的图片
     */
    private fun decodeBitmap(unzipDir: File, asset: LottieImageAsset): Bitmap {
        val localPath = unzipDir.absolutePath + File.separator + asset.dirName + File.separator + asset.fileName
        decodeAsync(localPath)
        return map.remove(localPath)?.let { return it } ?: decodeSync(localPath)
    }

    /**
     * 同步加载
     */
    private fun decodeSync(localPath: String) = try {
        BitmapFactory.decodeFile(localPath) ?: createDefaultBitmap()
    } catch (e: Throwable) {
        e.print()
        createDefaultBitmap()
    }

    /**
     * 创建默认图片
     */
    private fun createDefaultBitmap() = Bitmap.createBitmap(1, 1, Bitmap.Config.ALPHA_8)

    /**
     * 异步加载
     */
    private fun decodeAsync(localPath: String) {
        Observable.just(localPath)
                .subscribeOn(Schedulers.io())
                .map { BitmapFactory.decodeFile(localPath) }
                .subscribe({ map[localPath] = it }, { it.print() })
    }
}