package com.taobao.movie.android.commonui.utils

import android.app.Activity
import android.support.v4.view.ViewPager
import android.support.v7.widget.RecyclerView
import android.view.*
import android.view.ViewTreeObserver.OnDrawListener
import android.widget.HorizontalScrollView
import android.widget.ListView
import android.widget.ScrollView
import com.taobao.monitor.impl.data.DrawTimeCollector
import com.taobao.movie.android.commonui.component.BaseActivity
import com.taobao.movie.android.sdk.infrastructure.hack.MovieHacks
import com.taobao.movie.appinfo.util.LogUtil
import java.lang.reflect.Proxy
import kotlin.math.abs

/**
 * 策略：
 *  手指按下才开始算 fps，其他情况下不计算 fps
 *  手指没动的情况下得过滤掉
 *
 * @author yidong.yyd
 * @since 2019/11/8
 */
class FpsApmHelper(private val activity: Activity?) {

    companion object {
        /** Map<Activity, ActivityLifecycle.IPageLoadLifeCycle> */
        private var sPageLoadLifeCycleMap: Map<Activity, Any>? = null
        /** 比对次数 */
        const val MAX_TIME = 3
        /** 滑动比例 */
        const val SCROLL_RATE = 20
    }

    /** 手指的 X 坐标 */
    private var x = 0.0F
    /** 手指的 Y 坐标 */
    private var y = 0.0F
    /** 最后触摸的时间戳 */
    private var lastDownTimestamps = 0L
    /** 获取 APM 中的 [DrawTimeCollector] 对象  */
    private var drawTimeCollector: OnDrawListener? = null
    /** 是否 hook 失败了，失败的话就不要再去 hook 了  */
    private var isHooked = false
    /** 用户是否还可见 */
    private var isVisibleToUser = false
    /** 是否在滑动 */
    private var isScrolling = false
    /** 手指是否在屏幕上 */
    private var isFingerOnScreen = false
    /** 首页切 fragment 将状态变回去 */
    private var isHidden = false
    /** 最小滑动距离 */
    private val touchSlop: Int = ViewConfiguration.get(activity).scaledTouchSlop
    /** 次数 */
    private var slowMoveTimes = 0
    /** 对应控件的集合 */
    private val targetViewList = ArrayList<ViewGroup>()
    /** 防止重复添加拿来记录 */
    private val set: HashSet<View> = HashSet()

    /**
     * 与 onResume() 生命周期对齐
     * 在 resume 的时候不用把 drawTimeCollector 加上去，APM 会去加
     */
    fun onResume() {
        isVisibleToUser = true
        if (isHooked || activity == null || activity.window == null || activity.window.callback == null) {
            return
        }
        val isReflectSuccess = findDrawTimeCollector()
        if (activity is BaseActivity) {
            activity.onUTButtonClick("fps_effect_" + activity.utPageName, "reflect_effect", if (isReflectSuccess) "1" else "0")
        }
        if (!isReflectSuccess) {
            return
        }
        try {
            val callbackProxy = Proxy.newProxyInstance(activity.javaClass.classLoader,
                    arrayOf(Window.Callback::class.java),
                    MovieHacks.WindowCallbackProxy(activity.window.callback,
                            MovieHacks.WindowCallbackProxy.OnDispatchTouchEventInvokedListener(this::invoke)))
            if (callbackProxy is Window.Callback) {
                activity.window.callback = callbackProxy
            }
        } catch (e: Exception) {
            LogUtil.d("Exception", e.message ?: "Exception")
        } finally {
            isHooked = true
        }
        iterateView(activity?.window?.decorView)
        addListener()
    }

    /**
     * 与 onPause() 生命周期对齐
     */
    fun onPause() {
        isVisibleToUser = false
    }

    /**
     * 与 onDestroy() 生命周期对齐
     */
    fun onDestroy() {
        onPause()
        set.clear()
        targetViewList.clear()
    }

    /**
     * 停止滑动之后，将监听器移除掉，2s后再加回
     */
    fun scrollStateIdle() {
        isScrolling = false
        if (!isVisibleToUser) return
        removeListener()
    }

    /**
     * 正在滑动，可能是 dragging 或者 settling
     */
    fun scrollStateScrolling() {
        if (isScrolling) return
        isScrolling = true
        if (!isVisibleToUser) return
        addOnDrawListener()
    }

    /**
     * 界面隐藏和展示
     */
    fun onHiddenChanged(hidden: Boolean) {
        if (hidden) addOnDrawListener() else fastRemoveListener()
        isHidden = hidden
    }

    /**
     * 找 drawTimeCollector 监听器
     */
    private fun findDrawTimeCollector(): Boolean {
        findPageLoadLifeCycleMap()
        if (drawTimeCollector == null) drawTimeCollector = MovieHacks.getDrawTimeCollector(activity, sPageLoadLifeCycleMap)
        return sPageLoadLifeCycleMap != null && drawTimeCollector != null
    }

    /**
     * 找 Map<Activity, ActivityLifecycle.IPageLoadLifeCycle> 监听器
     */
    private fun findPageLoadLifeCycleMap() {
        sPageLoadLifeCycleMap = if (sPageLoadLifeCycleMap == null) {
            MovieHacks.getActivityLifecycleMap(activity)
        } else {
            sPageLoadLifeCycleMap
        }
    }

    /**
     * 代理触发时会回调这个
     */
    private fun invoke(motionEvent: MotionEvent) {
        isFingerOnScreen = motionEvent.action != MotionEvent.ACTION_UP && motionEvent.action != MotionEvent.ACTION_CANCEL
        // DOWN 的时候添加监听器
        if (motionEvent.action == MotionEvent.ACTION_DOWN) {
            addOnDrawListener()
        }
        // MOVE 事件
        detectMoveEvent(motionEvent)
        // UP 或者 CANCEL 的时候移掉监听器
        if (!isFingerOnScreen && !isScrolling) {
            removeListener()
        }
    }

    /**
     * 检测手势，如果手指在屏幕上但是没有产生滑动，那么不计算
     */
    private fun detectMoveEvent(motionEvent: MotionEvent) {
        if (motionEvent.action == MotionEvent.ACTION_MOVE) {
            var deltaX = abs(motionEvent.x - x)
            var deltaY = abs(motionEvent.y - y)
            var deltaTime = System.currentTimeMillis() - lastDownTimestamps
            if (deltaX < touchSlop && deltaY < touchSlop) {
                if ((deltaX == 0F && deltaY == 0F)
                        || (deltaY < deltaX && deltaX != 0F && (deltaTime / deltaX >= SCROLL_RATE))
                        || (deltaY > deltaX && deltaY != 0F && (deltaTime / deltaY >= SCROLL_RATE))) {
                    slowMoveTimes++
                }
                if (slowMoveTimes > MAX_TIME) {
                    slowMoveTimes = MAX_TIME
                    fastRemoveListener()
                }
            } else {
                slowMoveTimes--
                if (slowMoveTimes < 0) {
                    slowMoveTimes = 0
                    addOnDrawListener()
                }
            }
        }
        x = motionEvent.x
        y = motionEvent.y
        lastDownTimestamps = System.currentTimeMillis()
    }

    /**
     * 移除监听器
     */
    private fun removeListener() {
        if (isScrolling || isFingerOnScreen || isHidden) return
        fastRemoveListener()
        // 移除掉之后不再添加回去，当有 down 事件的时候再添加回去
    }

    /**
     * 直接移除
     */
    private fun fastRemoveListener() {
        if (isHidden) return
        activity?.window?.decorView?.viewTreeObserver?.removeOnDrawListener(drawTimeCollector)
    }

    /**
     * 添加监听器
     */
    private fun addOnDrawListener() {
        if (sPageLoadLifeCycleMap == null || drawTimeCollector == null || isHidden) return
        activity?.window?.decorView?.viewTreeObserver?.removeOnDrawListener(drawTimeCollector) // 先删除一遍，以防添加多次
        activity?.window?.decorView?.viewTreeObserver?.addOnDrawListener(drawTimeCollector)
    }

    /**
     * 找对应的View
     */
    private fun iterateView(view: View?) {
        if (view !is ViewGroup) return
        view.setOnHierarchyChangeListener(viewGroupOnHierarchyChangeListener)
        if (isTargetView(view) && !targetViewList.contains(view)) {
            targetViewList.add(view)
            addListener(view)
        }
        for (i in 0..view.childCount) {
            iterateView(view.getChildAt(i))
        }
    }

    /**
     * 是否是 RecyclerView、ViewPager、ScrollView、HorizontalScrollView、ListView
     */
    private fun isTargetView(view: View?): Boolean {
        return (view is RecyclerView || view is ViewPager || view is ScrollView || view is HorizontalScrollView || view is ListView)
    }

    /**
     * 监听子View被add remove 操作，然后重新计算
     */
    private val viewGroupOnHierarchyChangeListener = object : ViewGroup.OnHierarchyChangeListener {
        override fun onChildViewAdded(parent: View?, child: View?) {
            if (child == null) return
            if (isTargetView(parent)) {
                if (!set.contains(child)) {
                    set.add(child) // 防止重复添加
                    addListener(child)
                    iterateView(child)
                }
            } else {
                addListener(child)
                iterateView(child)
            }
        }

        override fun onChildViewRemoved(parent: View?, child: View?) {
        }
    }

    /**
     * 给 RecyclerView、ViewPager、ScrollView、HorizontalScrollView、ListView 添加监听器
     */
    private fun addListener() {
        targetViewList.forEach {
            if (it is RecyclerView) {
                it.removeOnScrollListener(recyclerViewOnScrollListener)
                it.addOnScrollListener(recyclerViewOnScrollListener)
            }
            if (it is ViewPager) {
                it.removeOnPageChangeListener(viewPagerOnPageChangeListener)
                it.addOnPageChangeListener(viewPagerOnPageChangeListener)
            }
            if (it is ScrollView || it is HorizontalScrollView) {
            }
            if (it is ListView) {
            }
        }
    }

    /**
     * 给 RecyclerView、ViewPager、ScrollView、HorizontalScrollView、ListView 添加监听器
     * @param view 对应的 View
     */
    private fun addListener(view: View) {
        if (view is RecyclerView) {
            view.removeOnScrollListener(recyclerViewOnScrollListener)
            view.addOnScrollListener(recyclerViewOnScrollListener)
        }
        if (view is ViewPager) {
            view.removeOnPageChangeListener(viewPagerOnPageChangeListener)
            view.addOnPageChangeListener(viewPagerOnPageChangeListener)
        }
        if (view is ScrollView || view is HorizontalScrollView) {
        }
        if (view is ListView) {
        }
    }

    /**
     * RecyclerView 的滑动监听
     */
    private val recyclerViewOnScrollListener = object : RecyclerView.OnScrollListener() {
        override fun onScrollStateChanged(recyclerView: RecyclerView, newState: Int) {
            if (newState == RecyclerView.SCROLL_STATE_IDLE) scrollStateIdle()
            else scrollStateScrolling()
        }
    }

    /**
     * ViewPager 的滑动监听
     */
    private val viewPagerOnPageChangeListener = object : ViewPager.OnPageChangeListener {
        override fun onPageScrollStateChanged(p0: Int) {
            if (p0 == ViewPager.SCROLL_STATE_IDLE) scrollStateIdle()
            else scrollStateScrolling()
        }

        override fun onPageScrolled(p0: Int, p1: Float, p2: Int) {
        }

        override fun onPageSelected(p0: Int) {
        }
    }
}